/*
 * ConversionHandler.cpp
 *
 *  Created on: 23. 2. 2014
 *      Author: tomas
 */

#include "ConversionHandler.h"

using namespace alib;
using namespace automaton;
using namespace grammar;
using namespace regexp;
using namespace sax;
using namespace std;

namespace conversions
{

ConversionHandler::ConversionHandler( list<Token> & tokens, const string & target, const string & algorithm, ostream & out ) :
    m_tokens( tokens ),
    m_source( parseFormalismFromTokens( ) ),
    m_target( parseFormalismFromString( target ) ),
    m_algorithm( parseAlgorithmFromString( algorithm ) ),
    m_out( out )
{

}

void ConversionHandler::convert( void )
{
    if( m_source == FINITE_AUTOMATON )
        convertFSM( );
    else if( m_source == REGULAR_EXPRESSION )
        convertRE( );
    else if( isGrammar( m_source ) )
        convertRG( );
    else
        throw AlibException( "ConversionHandler: Unknown source formalism." );
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertFSM( void )
{
    if( m_target == FINITE_AUTOMATON )
        throw AlibException( "ConversionHandler: No valid conversion from FSM to FSM." );
    else if( m_target == REGULAR_EXPRESSION )
        convertFSMtoRE( );
    else if( isGrammar( m_target ) )
        convertFSMtoRG( );
    else
        throw AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRE( void )
{
    if( m_target == FINITE_AUTOMATON )
        convertREtoFSM( );
    else if( m_target == REGULAR_EXPRESSION )
        throw AlibException( "ConversionHandler: No valid conversion from RE to RE." );
    else if( isGrammar( m_target ) )
        convertREtoRG( );
    else
        throw AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRG( void )
{
    if( m_target == FINITE_AUTOMATON )
        convertRGtoFSM( );
    else if( m_target == REGULAR_EXPRESSION )
        convertRGtoRE( );
    else if( isGrammar( m_target ) )
        convertRGtoRG( );
    else
        throw AlibException( "ConversionHandler: Unknown target formalism." );
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertFSMtoRE( void )
{
    const FSM fsm = AutomatonFactory::buildFSM( AutomatonParser::parse( m_tokens ) );

    IConversionRE* conv;

    switch( m_algorithm )
    {
    case BRZOZOWSKI_ALGEBRAIC:
        conv = new BrzozowskiAlgebraic( fsm );
        break;

    case STATE_ELIMINATION:
    default:
        conv = new StateElimination( fsm );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

void ConversionHandler::convertFSMtoRG( void )
{
    if( m_target == RIGHT_REGULAR_GRAMMAR )
        convertFSMtoRRG( );
    else if( m_target == LEFT_REGULAR_GRAMMAR )
        convertFSMtoLRG( );
    else
        throw AlibException( "ConversionHandler::Unknown target formalism" );
}

void ConversionHandler::convertFSMtoRRG( void )
{
    const FSM fsm = AutomatonFactory::buildFSM( AutomatonParser::parse( m_tokens ) );

    IConversionRRG* conv;

    switch( m_algorithm )
    {
    default:
        conv = new FAtoRRGConverter( fsm );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

void ConversionHandler::convertFSMtoLRG( void )
{
    const FSM fsm = AutomatonFactory::buildFSM( AutomatonParser::parse( m_tokens ) );

    IConversionLRG* conv;

    switch( m_algorithm )
    {
    default:
        conv = new FAtoLRGConverter( fsm );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertREtoFSM( void )
{
    const RegExp regexp = RegExpParser::parse( m_tokens );
    IConversionFSM* conv;

    switch( m_algorithm )
    {
    case BRZOZOWSKI_DERIVATION:
        conv = new Brzozowski( regexp );
        break;
    case THOMPSON_NFA:
        conv = new Thompson( regexp );
        break;
    case GLUSHKOV_NFA:
    default:
        conv = new Glushkov( regexp );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

void ConversionHandler::convertREtoRG( void )
{
    if( m_target == RIGHT_REGULAR_GRAMMAR )
        convertREtoRRG( );
    else if( m_target == LEFT_REGULAR_GRAMMAR )
        throw AlibException( "ConversionHandler:: RE to LRG is not implemented. Please convert to RRG and then to LRG." );
    else
        throw AlibException( "ConversionHandler::Unknown target formalism" );
}

void ConversionHandler::convertREtoRRG( void )
{
    const RegExp regexp = RegExpParser::parse( m_tokens );
    IConversionRRG* conv;

    switch( m_algorithm )
    {
    case BRZOZOWSKI_DERIVATION:
        conv = new BrzozowskiDerivationRRG( regexp );
        break;
    default:
        conv = new GlushkovRRG( regexp );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertRGtoFSM( void )
{
    if( m_source == LEFT_REGULAR_GRAMMAR )
        convertLRGtoFSM( );
    else if( m_source == RIGHT_REGULAR_GRAMMAR )
        convertRRGtoFSM( );
    else if( m_source == LEFT_LINEAR_GRAMMAR )
        throw AlibException( "ConversionHandler: Conversions from linear grammars are not yet implemented." );
    else if( m_source == RIGHT_LINEAR_GRAMMAR )
        throw AlibException( "ConversionHandler: Conversions from linear grammars are not yet implemented." );
    else
        throw AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRGtoRG( void )
{
    if( m_source == LEFT_REGULAR_GRAMMAR )
        convertLRGtoRRG( );
    else if( m_source == RIGHT_REGULAR_GRAMMAR )
        convertRRGtoLRG( );
    else if( m_source == LEFT_LINEAR_GRAMMAR )
        throw AlibException( "ConversionHandler: Conversions from linear grammars are not yet implemented." );
    else if( m_source == RIGHT_LINEAR_GRAMMAR )
        throw AlibException( "ConversionHandler: Conversions from linear grammars are not yet implemented." );
    else
        throw AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRGtoRE( void )
{
    if( m_source == LEFT_REGULAR_GRAMMAR )
        convertLRGtoRE( );
    else if( m_source == RIGHT_REGULAR_GRAMMAR )
        convertRRGtoRE( );
    else if( m_source == LEFT_LINEAR_GRAMMAR )
        throw AlibException( "ConversionHandler: Conversions from linear grammars are not yet implemented." );
    else if( m_source == RIGHT_LINEAR_GRAMMAR )
        throw AlibException( "ConversionHandler: Conversions from linear grammars are not yet implemented." );
    else
        throw AlibException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertLRGtoFSM( void )
{
    const LeftRegularGrammar lrg = GrammarFactory::buildLeftRegularGrammar( GrammarParser::parse( m_tokens ) );

    IConversionFSM* conv;

    switch( m_algorithm )
    {
    default:
        conv = new LRGtoFAConverter( lrg );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

void ConversionHandler::convertRRGtoFSM( void )
{
    const RightRegularGrammar rrg = GrammarFactory::buildRightRegularGrammar( GrammarParser::parse( m_tokens ) );

    IConversionFSM* conv;

    switch( m_algorithm )
    {
    default:
        conv = new RRGtoFAConverter( rrg );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

void ConversionHandler::convertLRGtoRE( void )
{
    const LeftRegularGrammar lrg = GrammarFactory::buildLeftRegularGrammar( GrammarParser::parse( m_tokens ) );

    IConversionRE* conv;

    switch( m_algorithm )
    {
    case BRZOZOWSKI_ALGEBRAIC:
    default:
        conv = new LRGAlgebraic( lrg );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

void ConversionHandler::convertRRGtoRE( void )
{
    const RightRegularGrammar rrg = GrammarFactory::buildRightRegularGrammar( GrammarParser::parse( m_tokens ) );

    IConversionRE* conv;

    switch( m_algorithm )
    {
    case BRZOZOWSKI_ALGEBRAIC:
    default:
        conv = new RRGAlgebraic( rrg );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

void ConversionHandler::convertLRGtoRRG( void )
{
    const LeftRegularGrammar lrg = GrammarFactory::buildLeftRegularGrammar( GrammarParser::parse( m_tokens ) );

    IConversionRRG* conv;

    switch( m_algorithm )
    {
    default:
        conv = new LeftToRightRegularGrammar( lrg );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

void ConversionHandler::convertRRGtoLRG( void )
{
    const RightRegularGrammar rrg = GrammarFactory::buildRightRegularGrammar( GrammarParser::parse( m_tokens ) );

    IConversionLRG* conv;

    switch( m_algorithm )
    {
    default:
        conv = new RightToLeftRegularGrammar( rrg );
        break;
    }

    conv->convert( ).toXML( m_out );
    delete conv;
}

// ----------------------------------------------------------------------------

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromString( const std::string & _target ) const
{
    string target;
    for( const auto & c : _target )
        target.append( 1, tolower( c ) );

    if( target == "fa" || target == "fsa" || target == "fsm" )
        return FINITE_AUTOMATON;

    if( target == "re" || target == "regexp" || target == "regex" )
        return REGULAR_EXPRESSION;

    if( target == "rrg" || target == "rg" || target == "grammar" )
        return RIGHT_REGULAR_GRAMMAR;

    if( target == "lrg" )
        return LEFT_REGULAR_GRAMMAR;

    throw AlibException( "ConversionHandler::Unknown target type" );
}

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromTokens( void ) const
{
    string xmlMark = m_tokens.front( ).getData( );

    if( xmlMark == "automaton" )
        return FINITE_AUTOMATON;

    if( xmlMark == "regexp" )
        return REGULAR_EXPRESSION;

    if( xmlMark == "grammar" )
    {
        bool ll, rl, lr, rr;
        ll = rl = lr = rr = true;

        list<Token> tokens_copy = m_tokens;
        UnknownGrammar grammar = GrammarParser::parse( tokens_copy );

        try
        {
            GrammarFactory::buildRightRegularGrammar( grammar );
        }
        catch( AlibException & e )
        {
            rr = false;
        }
        if( rr ) return RIGHT_REGULAR_GRAMMAR;

        try
        {
            GrammarFactory::buildLeftRegularGrammar( grammar );
        }
        catch( AlibException & e )
        {
            lr = false;
        }
        if( lr ) return LEFT_REGULAR_GRAMMAR;

        try
        {
            GrammarFactory::buildRightLinearGrammar( grammar );
        }
        catch( AlibException & e )
        {
            rl = false;
        }
        if( rl ) return RIGHT_LINEAR_GRAMMAR;

        try
        {
            GrammarFactory::buildLeftLinearGrammar( grammar );
        }
        catch( AlibException & e )
        {
            ll = false;
        }
        if( ll ) return LEFT_LINEAR_GRAMMAR;
    }

    throw AlibException( "ConversionHandler: Invalid input formalism." );
    // return UNKNOWN;
}

ConversionHandler::TAlgorithm ConversionHandler::parseAlgorithmFromString( const std::string & _algorithm ) const
{
    string algorithm;
    for( const auto & c : _algorithm )
        algorithm.append( 1, tolower( c ) );

    if( algorithm == "elimination" ) return STATE_ELIMINATION;
    if( algorithm == "algebraic" ) return BRZOZOWSKI_ALGEBRAIC;
    if( algorithm == "brzozowski" ) return BRZOZOWSKI_DERIVATION;
    if( algorithm == "glushkov" ) return GLUSHKOV_NFA;
    if( algorithm == "thompson" ) return THOMPSON_NFA;

    if( algorithm == "" ) return DEFAULT;

    throw AlibException( "ConversionHandler:: Invalid algorithm. See help." );
}

} /* namespace conversions */
