/*
 * StateElimination.cpp
 *
 *  Created on: 9. 2. 2014
 *      Author: tomas
 */

#include "StateElimination.h"

using namespace alib;
using namespace automaton;
using namespace regexp;
using namespace std;

namespace conversions
{

StateElimination::StateElimination( const FSM & automaton ) : m_automaton( automaton )
{

}

StateElimination::~StateElimination( void )
{
    deleteTransitions( );
}

void StateElimination::deleteTransitions( void )
{
    for( const auto & transition : m_MR.m_transitions )
        delete transition.m_regexp;
}

RegExp StateElimination::convert( void )
{
    if( m_automaton.getFinalStates( ).size( ) <= 0 )
    {
        RegExp ret;
        ret.setRegExp( new RegExpEmpty( ) );
        return ret;
    }

    m_MR = ExtendedNFA( );

    // steps 1 + 2
    constructExtendedNFA( );
    extendExtendedNFA( ); // yup, i am extending extended thing!

    // we are sure there is only 1 initial and 1 final state (follows from step2)
    const State & q0 = * m_MR.m_initialStates.begin( );
    const State & f = * m_MR.m_finalStates.begin( );

    // step 3 - iterating over all states except q0 and f, equivalent to the steps in Melichar 2.118
    for( auto it = m_MR.m_states.begin( ); it != m_MR.m_states.end( ); )
    {
        if( * it == q0 || * it == f )
            it ++;
        else
            eliminate( * it ++ );
    }

    // step 4
    Iteration * iter = new Iteration( );
    iter->setElement( transition( f, f ) );

    Concatenation* concat = new Concatenation( );
    concat->getElements( ).push_back( transition( q0, f ) );
    concat->getElements( ).push_back( iter );

    RegExpOptimize opt;

    RegExp ret;
    ret.setRegExp( opt.optimize( concat ) );

    delete concat;

    return ret;
}

void StateElimination::constructExtendedNFA( void )
{
    for( const auto & state : m_automaton.getStates( ) )
        m_MR.m_states.insert( state );

    for( const auto & symbol : m_automaton.getInputAlphabet( ) )
        m_MR.m_alphabet.insert( symbol );

    for( const auto & transition : m_automaton.getTransitions( ) )
        m_MR.m_transitions.insert( TransitionExtendedNFA( transition.getFrom( ), new RegExpSymbol( transition.getInput( ).getSymbol( ) ), transition.getTo( ) ) );

    for( const auto & state : m_automaton.getInitialStates( ) )
        m_MR.m_initialStates.insert( state );

    for( const auto & state : m_automaton.getFinalStates( ) )
        m_MR.m_finalStates.insert( state );
}

void StateElimination::extendExtendedNFA( void )
{
    if( isInSet( * m_MR.m_initialStates.begin( ), m_MR.m_finalStates ) || m_MR.transitionsToState( * m_MR.m_initialStates.begin( ) ).size( ) > 0 )
    {
        State q0 = m_MR.createUniqueState( "q0" );

        for( const auto & state : m_MR.m_initialStates )
            m_MR.m_transitions.insert( TransitionExtendedNFA( q0, new RegExpEpsilon( ), state ) );

        m_MR.m_initialStates.clear( );
        m_MR.m_initialStates.insert( q0 );
    }

    if( m_MR.m_finalStates.size( ) > 1 )
    {
        State f = m_MR.createUniqueState( "f" );

        for( const auto & state : m_MR.m_finalStates )
            m_MR.m_transitions.insert( TransitionExtendedNFA( state, new RegExpEpsilon( ), f ) );

        m_MR.m_finalStates.clear( );
        m_MR.m_finalStates.insert( f );
    }
}

void StateElimination::eliminate( const State & q )
{
    set<TransitionExtendedNFA> newTransitions;
    RegExpOptimize opt;

    for( const auto & p : m_MR.m_states )
    {
        for( const auto & r : m_MR.m_states )
        {
            if( p == q || r == q ) continue;

            // new transition function y(p,r) = y(p,r) + y(p,q)y(q,q)*y(q,r)
            Iteration* iter = new Iteration( );
            iter->setElement( transition( q, q ) );

            Concatenation* concat = new Concatenation( );
            concat->getElements( ).push_back( transition( p, q ) );
            concat->getElements( ).push_back( iter );
            concat->getElements( ).push_back( transition( q, r ) );

            Alternation* alt = new Alternation( );
            alt->getElements( ).push_back( transition( p, r ) );
            alt->getElements( ).push_back( concat );

            newTransitions.insert( TransitionExtendedNFA( p, opt.optimize( alt ), r ) );
            delete alt;
        }
    }

    m_MR.m_states.erase( q );

    deleteTransitions( );
    m_MR.m_transitions = newTransitions;
}

RegExpElement* StateElimination::transition( const State & p, const State & q ) const
{
    Alternation* alt = new Alternation( );

    alt->getElements( ).push_back( new RegExpEmpty( ) );

    for( const auto & transition : m_MR.m_transitions )
        if( transition.m_from == p && transition.m_to == q )
            alt->getElements( ).push_back( transition.m_regexp->clone( ) );

    return alt;
}

// ----------------------------------------------------------------------------

set<StateElimination::TransitionExtendedNFA> StateElimination::ExtendedNFA::transitionsToState( const State & toState ) const
{
    if( ! isInSet( toState, m_states ) )
        throw AlibException( "StateElimination::ExtendedNFA::transitionsToState - unknown state!" );

    set<TransitionExtendedNFA> transitions;

    for( const auto & transition : m_transitions )
        if( transition.m_to == toState )
            transitions.insert( transition );

    return transitions;
}

const State & StateElimination::ExtendedNFA::createUniqueState( const string & name )
{
    State uniqueState( name );
    if( ! isInSet( uniqueState, m_states ) )
    {
        m_states.insert( uniqueState );
        return * m_states.find( uniqueState );
    }

    int i = 0;
    while( i < INT_MAX )
    {
        State uniqueState( name + std::to_string( ++ i ) );

        if( ! isInSet( uniqueState, m_states ) )
        {
            m_states.insert( uniqueState );
            return * m_states.find( uniqueState );
        }
    }

     throw AlibException( "StateElimination::ExtendedNFA::createUniqueState - could not create unique state." );
}

// ----------------------------------------------------------------------------

StateElimination::TransitionExtendedNFA::TransitionExtendedNFA( const State & from, RegExpElement* regexp, const State & to ) : m_from( from ), m_to( to ), m_regexp( regexp )
{

}

bool StateElimination::TransitionExtendedNFA::operator<( const StateElimination::TransitionExtendedNFA & x ) const
{
    if( m_from != x.m_from )
        return m_from < x.m_from;
    else if( m_regexp != x.m_regexp )
        return m_regexp < x.m_regexp;
    else
        return m_to < x.m_to;
}

} /* namespace conversions */
