/*
 * Thompson.cpp
 *
 *  Created on: 11. 1. 2014
 *      Author: tomas
 */
#include "Thompson.h"

using namespace alib;
using namespace automaton;
using namespace regexp;

namespace conversions
{

Thompson::Thompson( const RegExp & re ) : m_re( re )
{

}

Thompson::~Thompson( void )
{

}


FSM Thompson::convert( void )
{
    m_fsm = FSM( );
    m_stateId = 0;

    for( const auto & symbol : m_re.getAlphabet( ) )
        m_fsm.addInputSymbol( symbol.getSymbol( ) );

    SubexpressionTails st = processRegExpNode( m_re.getRegExp( ) );

    m_fsm.addInitialState( st.m_head );
    m_fsm.addFinalState( st.m_tail );

    return m_fsm;
}

Thompson::SubexpressionTails Thompson::processRegExpNode( const RegExpElement * node )
{
    const Alternation* alternation = dynamic_cast<const Alternation*>( node );
    const Concatenation* concatenation = dynamic_cast<const Concatenation*>( node );
    const Iteration* iteration = dynamic_cast<const Iteration*>( node );
    const RegExpSymbol* symbol = dynamic_cast<const RegExpSymbol*>( node );
    const RegExpEmpty* empty = dynamic_cast<const RegExpEmpty*>( node );
    const RegExpEpsilon* eps = dynamic_cast<const RegExpEpsilon*>( node );

    if( alternation )
        return processRegExpNode( alternation );
    else if( concatenation )
        return processRegExpNode( concatenation );
    else if( iteration )
        return processRegExpNode( iteration );
    else if( symbol )
        return processRegExpNode( symbol );
    else if( eps )
        return processRegExpNode( eps );
    else if( empty )
        return processRegExpNode( empty );

     throw AlibException( "Thompson::process - invalid RegExpElement node." );
}

Thompson::SubexpressionTails Thompson::processRegExpNode( const Iteration * node )
{
    State head = m_fsm.createUniqueState( toBase26( m_stateId    ) + "0", true );
    State tail = m_fsm.createUniqueState( toBase26( m_stateId ++ ) + "1", true );

    SubexpressionTails st = processRegExpNode( node->getElement( ) );

    m_fsm.addTransition( head, Symbol( "" ), st.m_head );
    m_fsm.addTransition( head, Symbol( "" ), tail );
    m_fsm.addTransition( st.m_tail, Symbol( "" ), tail );
    m_fsm.addTransition( st.m_tail, Symbol( "" ), st.m_head );

    return SubexpressionTails( head, tail );
}

Thompson::SubexpressionTails Thompson::processRegExpNode( const Alternation * node )
{
    State head = m_fsm.createUniqueState( toBase26( m_stateId    ) + "0", true );
    State tail = m_fsm.createUniqueState( toBase26( m_stateId ++ ) + "1", true );

    for( const auto & element : node->getElements( ) )
    {
        SubexpressionTails st = processRegExpNode( element );

        m_fsm.addTransition( head, Symbol( "" ), st.m_head );
        m_fsm.addTransition( st.m_tail, Symbol( "" ), tail );
    }

    return SubexpressionTails( head, tail );
}

Thompson::SubexpressionTails Thompson::processRegExpNode( const Concatenation * node )
{
    vector<SubexpressionTails> st;
    for( const auto & element : node->getElements( ) )
        st.push_back( processRegExpNode( element ) );

    for( size_t i = 1; i < st.size( ); i ++ )
        m_fsm.addTransition( st[ i - 1 ].m_tail, Symbol( "" ), st[ i ].m_head );

    return SubexpressionTails( st[ 0 ].m_head, st[ st.size( ) - 1 ].m_tail );
}

Thompson::SubexpressionTails Thompson::processRegExpNode( const RegExpSymbol * node )
{
    Symbol symb( node->getSymbol( ) );
    State head = m_fsm.createUniqueState( toBase26( m_stateId    ) + "0", true );
    State tail = m_fsm.createUniqueState( toBase26( m_stateId ++ ) + "1", true );

    m_fsm.addTransition( head, symb, tail );

    return SubexpressionTails( head, tail );
}

Thompson::SubexpressionTails Thompson::processRegExpNode( const RegExpEpsilon * node )
{
    Symbol symb( "" );
    State head = m_fsm.createUniqueState( toBase26( m_stateId    ) + "0", true );
    State tail = m_fsm.createUniqueState( toBase26( m_stateId ++ ) + "1", true );

    m_fsm.addTransition( head, symb, tail );

    return SubexpressionTails( head, tail );
}

Thompson::SubexpressionTails Thompson::processRegExpNode( const RegExpEmpty * node )
{
    State head = m_fsm.createUniqueState( toBase26( m_stateId    ) + "0", true );
    State tail = m_fsm.createUniqueState( toBase26( m_stateId ++ ) + "1", true );

    return SubexpressionTails( head, tail );
}

// ----------------------------------------------------------------------------

Thompson::SubexpressionTails::SubexpressionTails( const State & head, const State & tail ) : m_head( head ), m_tail ( tail )
{

}

} /* namespace conversions */
